#!/bin/bash
echo "===================================================="
echo "AWS EC2 Distributed Cluster Setup Script"
echo "Create a cluster on AWS EC2 Ubuntu20.04 and set up the environment for MPI"
echo "For EC2 Spot instances, please input 'spot' as 1st Argument with the number of instances as 2nd Argument; "
echo "For EC2 ON Demand instances, please input 'demand' as 1st Argument with the number of instances as 2nd Argument;"
echo "One-to-Many: SSH Connection between Master and WORKERS without key"
echo "Author: Haibin Guan"
echo "===================================================="

echo "===================================================="
echo "set up for AWS EC2 $1 instances"
echo "===================================================="
echo "Remove the previous information of instances under the /instance_info"

#### For example:
#### ====================================================
#### For Spot instances:
#### bash setup_onetomany.sh spot 51 r5dn.xlarge c5n.large
#### ====================================================
#### For On Demand instances:
#### bash setup_onetomany.sh demand 51


rm instance_info/*

#GLobal Variables
PEM_FILE=local_alloy.pem
LOCAL_PEM=keypair/$PEM_FILE
PASS=111          #
USER=pedro       #
TEST_FOLDER=scripts/
SCRIPT_FOLDER=../*.py
NN_FOLDER=../nn
mnist_file=../artificial-data/mnist/train.csv
sourcefile_path=scripts/sourcefile
USER_ACCESS=ubuntu #based on the AMI you picked
SSH_ATTEMPTS=5    #
SLEEPTIME=3
MASTER_FOLDER=master_files/*      #

## command to get the instance_id of running instances before running this setup script
##-------------------------------------------------------
if [[ $1 == 'spot' ]];then
NUM_NODES=$2
MASTER_TYPE=$3
WORKER_TYPE=$4

if [[ $MASTER_TYPE == $WORKER_TYPE ]];then
  aws ec2 describe-instances --profile psoto --filters "Name=instance-state-code,Values=16" \
  --query 'Reservations[].Instances[].InstanceId' > instance_info/id_instances.json
else
  aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$MASTER_TYPE" "Name=instance-state-code,Values=16" \
  --query 'Reservations[].Instances[].InstanceId' > instance_info/id_instances.json
  aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$WORKER_TYPE" "Name=instance-state-code,Values=16" \
  --query 'Reservations[].Instances[].InstanceId' >> instance_info/id_instances.json
fi

id_instances=($(jq -r '.[]' instance_info/id_instances.json) )

#NUM_NODES=($(jq -r '.[]' length instance_info/id_instances.json) ) #number of running instances



elif [[ $1 == 'demand' ]];then

NUM_NODES=$2 # take the first argument
echo "===================================================="
echo "Creating $NUM_NODES AWS EC2 instances for our distributed cluster"
echo "===================================================="
aws ec2 run-instances --profile psoto --image-id ami-0fb653ca2d3203ac1 \
--security-group-ids sg-01c88ccbc51494cb0 \
--count $NUM_NODES --instance-type t2.micro \
--key-name local_alloy --subnet-id  subnet-0d9d1f70fac6e1dca \
--associate-public-ip-address --query 'Instances[*].InstanceId' \
>> instance_info/id_instances.json
echo "DONE and Saved instances info on instance_info/id_instances.json"
id_instances=( $(jq -r '.[]' instance_info/id_instances.json) )

id_inst_params=""
for each in "${id_instances[@]}"
do
id_inst_params="$id_inst_params $each"
done

echo "===================================================="
echo "Make sure the status of all of the instances are RUNNING"
echo "===================================================="
running_nodes=0
while [[ $running_nodes != $NUM_NODES ]]; do

aws ec2 describe-instance-status --profile psoto --instance-ids $id_inst_params \
--query "InstanceStatuses[*].InstanceState.Name" \
> instance_info/status_instances.json

status_instances=( $(jq -r '.[]' instance_info/status_instances.json) )

if [[ ${#status_instances[@]} = 0 ]]; then
  sleep $SLEEPTIME
  continue
fi  #indicates the end of the inner if statement

  running_nodes=0
  for status in "${status_instances[@]}"; do
  if [[ status != "running" ]]; then
  running_nodes=$((running_nodes+1))
  fi
  done
  sleep $SLEEPTIME
done

echo "DONE, All Running!"
else
  echo "===================================================="
  echo "Wrong Argument! First argument either be 'spot' or 'demand' !"
  echo "===================================================="
  exit
fi

echo "===================================================="
echo "Getting the Public and Private IPs"
echo "Saved IPs on instance_info/public_ip_list.json and instance_info/ip_private_list.json"
echo "===================================================="

if [[ $MASTER_TYPE == $WORKER_TYPE ]];then
aws ec2 describe-instances --profile psoto --filters "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PublicIpAddress' \
>> instance_info/public_ip_list.json
aws ec2 describe-instances --profile psoto --filters "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PrivateIpAddress' \
>> instance_info/ip_private_list.json

else
aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$MASTER_TYPE" "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PublicIpAddress' \
>> instance_info/public_ip_list.json
aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$WORKER_TYPE" "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PublicIpAddress' \
>> instance_info/public_ip_list.json

aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$MASTER_TYPE" "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PrivateIpAddress' \
>> instance_info/ip_private_list.json
aws ec2 describe-instances --profile psoto --filters "Name=instance-type,Values=$WORKER_TYPE" "Name=instance-state-code,Values=16" \
--query 'Reservations[].Instances[].PrivateIpAddress' \
>> instance_info/ip_private_list.json
fi

## list ip instances
public_ip_list=( $(jq -r '.[]' instance_info/public_ip_list.json) )
ip_private_list=( $(jq -r '.[]' instance_info/ip_private_list.json) )

## setting MASTER
MASTER=${public_ip_list[0]}
MASTER_PRIVATE_IP=${ip_private_list[0]}
echo "===================================================="
echo "Checking SSH connections on instances"
echo "===================================================="
for pub_ip in "${public_ip_list[@]}"
do
ssh -oStrictHostKeyChecking=no -oConnectionAttempts=$SSH_ATTEMPTS -i $LOCAL_PEM $USER_ACCESS@$pub_ip "exit;"
echo "$pub_ip is READY!"
done
echo "NODES ARE READY TO BE SET UP!"


echo "===================================================="
echo "Configure the MASTER::$MASTER NODE"
echo "===================================================="
scp -i $LOCAL_PEM -r $NN_FOLDER $USER_ACCESS@$MASTER:~/
scp -i $LOCAL_PEM -r $TEST_FOLDER $USER_ACCESS@$MASTER:~/
scp -i $LOCAL_PEM -r $SCRIPT_FOLDER $USER_ACCESS@$MASTER:~/
echo "===================================================="
echo "Sending the AMAZON PEM and pushout.sh to the MASTER::$MASTER node "
echo "===================================================="
scp -i $LOCAL_PEM $LOCAL_PEM $USER_ACCESS@$MASTER:~/
scp -i $LOCAL_PEM -r $MASTER_FOLDER $USER_ACCESS@$MASTER:~/

##config MASTER [cp al posto di cat]
master_conf="sudo useradd -s /bin/bash -m -d /home/$USER -g root $USER; \
echo -e \"$PASS\n$PASS\n\" | sudo passwd $USER;\
chmod 700 .ssh;\
chmod 600 .ssh/authorized_keys;\
sudo apt-get update;\
sudo apt-get install -y -q python3-dev python3-pip mpich python3-mpi4py; \
pip3 install -q numpy scipy scikit-learn matplotlib pandas;\
mkdir figures;\
mkdir artificial-data;\
mkdir artificial-data/mnist;\
source $sourcefile_path;"

scp -i $LOCAL_PEM mnist_file $USER_ACCESS@$MASTER:artificial-data/mnist/

ssh -oStrictHostKeyChecking=no -i $LOCAL_PEM $USER_ACCESS@$MASTER "$master_conf"
echo "DONE WITH Configuring the MASTER::$MASTER "
echo "===================================================="

echo "===================================================="
echo "Configurethe WORKERS NODES                           "
echo "===================================================="
for (( i=1; i<$NUM_NODES; i++ ))
do
curr_WORKER_public_ip=${public_ip_list[$i]}
scp -i $LOCAL_PEM -r $TEST_FOLDER $USER_ACCESS@$curr_WORKER_public_ip:~/
scp -i $LOCAL_PEM -r $SCRIPT_FOLDER $USER_ACCESS@$curr_WORKER_public_ip:~/
scp -i $LOCAL_PEM -r $NN_FOLDER $USER_ACCESS@$curr_WORKER_public_ip:~/
scp -i $LOCAL_PEM $LOCAL_PEM $USER_ACCESS@$curr_WORKER_public_ip:~/

WORKER_conf="sudo useradd -s /bin/bash -m -d /home/$USER -g root $USER;\
echo -e \"$PASS\n$PASS\n\" | sudo passwd $USER;\
chmod 700 .ssh;\
chmod 600 .ssh/authorized_keys;\
sudo apt-get update; \
sudo apt-get install -y -q python3-dev python3-pip mpich python3-mpi4py;\
pip3 install -q numpy scipy scikit-learn matplotlib pandas;\
source $sourcefile_path;"

ssh -oStrictHostKeyChecking=no -i $LOCAL_PEM $USER_ACCESS@$curr_WORKER_public_ip "$WORKER_conf"
done

echo "DONE WITH Configuring the WORKERS "
echo "===================================================="


echo "===================================================="
echo "Create hostsfile ..."
echo "===================================================="
rm hostsfile
for private_ip in "${ip_private_list[@]}"
do
echo "$private_ip" >> hostsfile
done
echo "Hostsfile has been created!"
echo "===================================================="

echo "===================================================="
echo "Send the hostsfile and add the private IPs in /etc/hosts for all nodes"
echo "===================================================="
set_hosts="printf \"\n#AWS Build Cluster Script -- ip private nodes\n\" | sudo tee -a /etc/hosts; "
set_hosts=$set_hosts"printf \"${ip_private_list[0]}   MASTER\n\" | sudo tee -a /etc/hosts; "
for (( i=1; i<$NUM_NODES; i++ ))
do
set_hosts=$set_hosts"printf \"${ip_private_list[$i]}   NODE_$i\n\" | sudo tee -a /etc/hosts; "
done
for node in "${public_ip_list[@]}"
do
scp -i $LOCAL_PEM hostsfile $USER_ACCESS@$node:~/
ssh -i $LOCAL_PEM $USER_ACCESS@$node "$set_hosts"
done

echo "===================================================="
echo "===================================================="
echo "Generate the key file and send the id_rsa.pub to the rest of nodes "
echo "cat id_rsa.pub >> .ssh/authorized_keys "
echo "===================================================="
key_conf="ssh-keygen -t rsa -P \"\" -f .ssh/id_rsa;\
chmod 600 .ssh/id_rsa.pub;\
chmod 600 .ssh/id_rsa;\
sudo bash -c 'cat .ssh/id_rsa.pub >> .ssh/authorized_keys';"

ssh -oStrictHostKeyChecking=no -i $LOCAL_PEM $USER_ACCESS@$MASTER "$key_conf"
for (( j=1; j<$NUM_NODES; j++ ))
do
curr_recv_public_ip=${public_ip_list[$j]}
echo "$MASTER send id_ras $curr_recv_public_ip"
echo "===================================================="
send_rsa="sudo scp -oStrictHostKeyChecking=no -i $PEM_FILE .ssh/id_rsa.pub $USER_ACCESS@$curr_recv_public_ip:"
ssh -i $LOCAL_PEM $USER_ACCESS@$MASTER "$send_rsa"
pem_conf="sudo bash -c 'cat id_rsa.pub >> .ssh/authorized_keys';"
ssh -i $LOCAL_PEM $USER_ACCESS@$curr_recv_public_ip "$pem_conf"
done


send_rsa="sudo scp -oStrictHostKeyChecking=no -i $PEM_FILE .ssh/id_rsa.pub $USER_ACCESS@$MASTER:"
for (( i=1; i<$NUM_NODES; i++ ))
do
curr_send_public_ip=${public_ip_list[$i]}
ssh -oStrictHostKeyChecking=no -i $LOCAL_PEM $USER_ACCESS@$curr_send_public_ip "$key_conf"
echo "$curr_send_public_ip send id_ras $curr_recv_public_ip"
echo "===================================================="
ssh -i $LOCAL_PEM $USER_ACCESS@$curr_send_public_ip "$send_rsa"
ssh -i $LOCAL_PEM $USER_ACCESS@$MASTER "$pem_conf"
done

echo "DONE"
echo "===================================================="

echo "===================================================="
echo "Trying to ssh connect without keys                  "
echo "===================================================="
curr_send_public_ip=$MASTER
for (( j=1; j<$NUM_NODES; j++ ))
do
curr_recv_private_ip=${ip_private_list[$j]}
echo "===================================================="
echo "YAY!!!! $curr_send_public_ip ssh connect $curr_recv_private_ip without key"
echo "===================================================="
ssh_conf="ssh -oStrictHostKeyChecking=no $curr_recv_private_ip "exit";"
ssh -t -i $LOCAL_PEM $USER_ACCESS@$curr_send_public_ip "$ssh_conf"
done

curr_recv_private_ip=$MASTER_PRIVATE_IP
for (( i=1; i<$NUM_NODES; i++ ))
  do
curr_send_public_ip=${public_ip_list[$i]}
echo "===================================================="
echo "YAY!!!! $curr_send_public_ip ssh connect $curr_recv_private_ip without key"
echo "===================================================="
ssh_conf="ssh -oStrictHostKeyChecking=no $curr_recv_private_ip "exit";"
ssh -t -i $LOCAL_PEM $USER_ACCESS@$curr_send_public_ip "$ssh_conf"
done

echo "Done With SSH connection without keys               "
echo "===================================================="

echo "===================================================="
echo "Congradulations!!! The AWS Cluster for MPI is READY!"
echo "===================================================="
echo "On each instance there is a user with : "
echo "USERNAME:$USER - PASSWORD:$PASS"
echo -e "MASTER \tPRIVATE_IP=$MASTER_PRIVATE_IP \tPUBLIC_IP=$MASTER"

for (( i=1; i<$NUM_NODES; i++ ))
do
curr_private_WORKER_ip=${ip_private_list[$i]}
curr_public_WORKER_ip=${public_ip_list[$i]}
echo -e "WORKER $i\tPRIVATE_IP=$curr_private_WORKER_ip \tPUBLIC_IP=$curr_public_WORKER_ip"
done
