# only use custom functions and raw python code
from mpi4py import MPI
import numpy as np
import argparse
from sklearn.datasets import make_classification, make_regression

# "copy" gradient coding to this file
parser = argparse.ArgumentParser()

parser.add_argument("--num-workers", "-nw", default=4, type=int)

args = parser.parse_args()


def experiment(
    is_regr: bool = True,
    n_samples: int = 100,
    n_features: int = 2,
    n_workers: int = 4,
    n_epochs: int = 10,
):

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    # assert size <= n_workers - 1

    rows_per_worker = n_samples // n_workers

    # generate data

    if rank == 0:
        if is_regr:
            X, y = make_regression(n_samples, n_features)
        else:
            X, y = make_classification(n_samples, n_features)

        weights = np.random.randn(n_features)

        for i in range(1, n_workers + 1):
            comm.Send(
                X[(i - 1) * rows_per_worker : i * rows_per_worker],
                dest=i,
                tag=11,
            )
            comm.Send(
                y[(i - 1) * rows_per_worker : i * rows_per_worker],
                dest=i,
                tag=12,
            )

    else:
        X_on_worker = np.empty((rows_per_worker, n_features))
        y_on_worker = np.empty((rows_per_worker,))
        comm.Recv(X_on_worker, source=0, tag=11)
        comm.Recv(y_on_worker, source=0, tag=12)

    # comm.Barrier()
    for epoch in range(n_epochs):
        if rank == 0:
            print(f"Epoch: {epoch+1}")

            if (epoch + 1) % 10 == 0:
                print(f"{np.mean((0.5*np.dot(X, weights)-y)**2)}")

            g = np.empty((n_features,))
            for i in range(1, n_workers + 1):
                # send weights to workers
                comm.Send(
                    weights,
                    dest=i,
                    tag=21,
                )

            for i in range(1, n_workers + 1):
                comm.Recv(
                    g,
                    source=i,
                    tag=31,
                )

                weights -= 0.01 * g
        else:
            weights_on_worker = np.empty((n_features,))
            comm.Recv(weights_on_worker, source=0, tag=21)

            output = np.dot(X_on_worker, weights_on_worker)  # np.maximum(, 0)

            gradient = X_on_worker.T.dot(
                (output - y_on_worker) / output.shape[0]
            )
            comm.Send(gradient, dest=0, tag=31)
    # comm.Barrier()
    # MPI.Finalize()

    if rank == 0:
        import matplotlib.pyplot as plt

        y_pred = np.dot(X, weights)
        plt.plot(y_pred, X)
        plt.plot(y, X)

        plt.show()


experiment(
    n_samples=200, n_features=1, n_workers=args.num_workers, n_epochs=100
)
