from __future__ import print_function

import argparse
import os
import random
import sys

import numpy as np
import pandas as pd

from util import (
    generate_random_binvec,
    save_matrix,
    save_vector,
    generate_random_matrix_normal,
)


def generate_data(partitions, out_dir):
    # partitions = 1
    # make sure the number of rows is a multiple of number of total partitions
    assert n_rows % partitions == 0

    # num workers, and number of rows per worker
    n_rows_per_worker = n_rows // partitions

    # Create output directory
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    print(
        "Generating Partitioned Matrix Of size %d x %d for a total of %d partitions"
        % (n_rows, n_cols, partitions)
    )
    print(
        ">>> Each worker gets a matrix of %d x %d doubles, %.2f MB each"
        % (
            n_rows_per_worker,
            n_cols,
            (n_rows_per_worker * n_cols * 8) / 1000000.0,
        )
    )

    random_beta = generate_random_binvec(n_cols, n_classes).T
    alpha = np.random.beta(1, 2) + np.random.choice([-1, 1], 1)[0]  # 1.5
    mu = [
        np.multiply(alpha / n_cols, random_beta).sum(-1)
        for _ in range(n_classes)
    ]  # .sum(-1)
    # mu2 = np.multiply(-alpha / n_cols, random_beta).sum(-1)
    label_vector = np.zeros(n_rows)

    for i in range(1, partitions + 1):

        random_matrix = generate_random_matrix_normal(
            mu, n_rows_per_worker, n_cols, n_classes
        )
        save_matrix(random_matrix, out_dir + str(i) + ".dat")

        prob_vals = np.divide(
            1.0, np.add(1, np.exp(-np.dot(random_matrix, random_beta)))
        )

        prob_vals = prob_vals / prob_vals.sum(1).reshape((-1, 1))

        label_vector[
            (i - 1) * int(n_rows_per_worker) : i * int(n_rows_per_worker)
        ] = [
            np.random.multinomial(
                1,
                pv,
            ).argmax()
            for pv in prob_vals
        ]
        # label_j = sum X_Ij, where X_Ij = 1 with prob prob_vals[j]

        print("\t >>> Done with partition %d" % (i))

    save_vector(label_vector, out_dir + "label.dat")

    random_matrix_test = generate_random_matrix_normal(
        mu, int(0.2 * n_rows), n_cols, n_classes
    )
    prob_vals = np.divide(
        1.0, np.add(1, np.exp(-np.dot(random_matrix_test, random_beta)))
    )
    prob_vals = prob_vals / prob_vals.sum(1).reshape((-1, 1))

    label_vector_test = np.array(
        [
            np.random.multinomial(
                1,
                pv,
            ).argmax()
            for pv in prob_vals
        ]
    )

    assert random_matrix_test.shape[0] == label_vector_test.shape[0]
    save_matrix(random_matrix_test, out_dir + "test_data.dat")
    save_vector(label_vector_test, out_dir + "label_test.dat")
    print("\t >>> Done with Test data")


# ---------- code for call from Makefile starts here
# if len(sys.argv) != 8:
#     print(
#         "Usage: python generate_data.py n_procs n_rows n_cols output_dir n_stragglers n_partitions partial_coded"
#     )
#     sys.exit(0)

parser = argparse.ArgumentParser()

parser.add_argument("--n-procs", type=int)
parser.add_argument("--n-rows", type=int)
parser.add_argument("--n-cols", type=int)
parser.add_argument("--output-dir", type=str)
parser.add_argument("--n-stragglers", type=int)
parser.add_argument("--n-partitions", type=int)
parser.add_argument("--partial_coded", type=str)

parser.add_argument("--num-classes", type=int)

args = parser.parse_args()

n_procs = args.n_procs
n_rows = args.n_rows
n_cols = args.n_cols
output_dir = args.output_dir
n_stragglers = args.n_stragglers
n_partitions = args.n_partitions
partial_coded = args.partial_coded == "True"

n_classes = args.num_classes

# np.random.seed(0)
# (
#     n_procs,
#     n_rows,
#     n_cols,
#     output_dir,
#     n_stragglers,
#     n_partitions,
#     partial_coded,
# ) = [x for x in sys.argv[1:]]
# n_procs, n_cols, n_rows, n_stragglers, n_partitions, partial_coded = (
#     int(n_procs),
#     int(n_cols),
#     int(n_rows),
#     int(n_stragglers),
#     int(n_partitions),
#     int(partial_coded),
# )

print(n_procs, n_cols, n_rows, n_stragglers, n_partitions, partial_coded)
output_dir = output_dir + "/" if not output_dir[-1] == "/" else output_dir

if not partial_coded:
    output_dir = (
        output_dir
        + "artificial-data/"
        + str(n_rows)
        + "x"
        + str(n_cols)
        + "/"
        + str(n_procs - 1)
        + "/"
    )
else:
    output_dir = (
        output_dir
        + "artificial-data/"
        + str(n_rows)
        + "x"
        + str(n_cols)
        + "/partial/"
        + str((n_procs - 1) * (n_partitions - n_stragglers))
        + "/"
    )

# generate data
if not partial_coded:
    generate_data(n_procs - 1, output_dir)
else:
    # generate data for partial replication
    generate_data((n_procs - 1) * (n_partitions - n_stragglers), output_dir)


mnist_train = pd.read_csv("artificial-data/mnist/train.csv")
output_dir = "artificial-data/mnist/"
from sklearn.model_selection import train_test_split

x_train, x_val = train_test_split(
    mnist_train, test_size=2000, random_state=42, stratify=mnist_train["label"]
)

partitions = n_procs - 1
train_label = x_train["label"]
val_label = x_val["label"]

x_val.drop(["label"], axis=1, inplace=True)
x_train.drop(["label"], axis=1, inplace=True)


for idx, partition in enumerate(np.array_split(x_train, partitions)):
    partition.to_csv(
        output_dir + str(idx + 1) + ".dat", sep=" ", index=False, header=False
    )

x_val.to_csv(output_dir + "test_data.dat", sep=" ", index=False, header=False)

train_label.to_csv(
    output_dir + "label.dat", sep=" ", index=False, header=False
)
val_label.to_csv(
    output_dir + "label_test.dat", sep=" ", index=False, header=False
)

print("Data Generation Finished.")
