from __future__ import print_function
import enum
import sys
import random
import os
import numpy as np
import itertools
import scipy.special as sp
from scipy.sparse import csr_matrix

# ---- Data generation, saving, loading and modification routines


def load_data(input_file):
    mydata = np.loadtxt(input_file, dtype=float)
    return mydata


def save_sparse_csr(filename, array):
    np.savez(
        filename,
        data=array.data,
        indices=array.indices,
        indptr=array.indptr,
        shape=array.shape,
    )


def load_sparse_csr(filename):
    loader = np.load(filename + ".npz")
    return csr_matrix(
        (loader["data"], loader["indices"], loader["indptr"]),
        shape=loader["shape"],
    )


def save_matrix(m, output):
    f = open(output, "w")
    for i in range(len(m)):
        print(" ".join([str(x) for x in m[i]]), file=f)
    f.close()


def save_vector(m, output):
    f = open(output, "w")
    for i in range(len(m)):
        print("%5.3f" % (m[i]) + " ", file=f)
    f.close()


# generates a random matrix representing samples from a two-component GMM with identity covariance
def generate_random_matrix_normal(mu1, n_rows, n_cols, n_classes):
    ctrmu_i = []
    for i in range(n_classes - 1):
        ctrmu_i.append(
            np.random.choice(list(range(n_rows - sum(ctrmu_i))), replace=False)
        )
    ctrmu_i.append(n_rows - sum(ctrmu_i))
    print(ctrmu_i, sum(ctrmu_i), n_rows)
    # ctrmu2 = np.random.binomial(n_rows, 0.5)
    # ctrmu1 = n_rows - ctrmu2
    mfac = 10 / np.sqrt(n_cols)

    for i, each in enumerate(ctrmu_i):
        print(
            np.add(
                mfac * np.random.standard_normal((int(each), n_cols)), mu1[i]
            ).shape
        )

    return np.concatenate(
        tuple(
            [
                np.add(
                    mfac * np.random.standard_normal((int(each), n_cols)),
                    mu1[i],
                )
                for i, each in enumerate(ctrmu_i)
                # np.add(
                #     mfac * np.random.standard_normal((ctrmu2, n_cols)), mu2
                # ),
            ]
        )
    )


# generates a vector of random labels, each entry only has value -1 or 1
def generate_random_binvec(n, num_class=None):
    if num_class:
        return np.array(
            [
                [np.random.randint(2) * 2 - 1 for x in range(n)]
                for _ in range(num_class)
            ]
        )
    return np.array([np.random.randint(2) * 2 - 1 for x in range(n)])


def interactionTermsAmazon(data, degree, hash=hash):
    new_data = []
    m, n = data.shape
    for indicies in itertools.combinations(range(n), degree):
        if not (5 in indicies and 7 in indicies) and not (
            2 in indicies and 3 in indicies
        ):
            new_data.append([hash(tuple(v)) for v in data[:, indicies]])
    return np.array(new_data).T


# ---- Other routines


def unique_rows(a):
    b = np.ascontiguousarray(a).view(
        np.dtype((np.void, a.dtype.itemsize * a.shape[1]))
    )
    _, idx = np.unique(b, return_index=True)
    return a[idx]


def getB(n_workers, n_stragglers):
    Htemp = np.random.normal(0, 1, [n_stragglers, n_workers - 1])
    H = np.vstack([Htemp.T, -np.sum(Htemp, axis=1)]).T

    Ssets = np.zeros([n_workers, n_stragglers + 1])

    for i in range(n_workers):
        Ssets[i, :] = np.arange(i, i + n_stragglers + 1)
    Ssets = Ssets.astype(int)
    Ssets = Ssets % n_workers
    B = np.zeros([n_workers, n_workers])
    for i in range(n_workers):
        B[i, Ssets[i, 0]] = 1
        vtemp = -np.linalg.solve(
            H[:, np.array(Ssets[i, 1:])], H[:, Ssets[i, 0]]
        )
        ctr = 0
        for j in Ssets[i, 1:]:
            B[i, j] = vtemp[ctr]
            ctr += 1

    return B


def getA(B, n_workers, n_stragglers):
    # S=np.array(list(itertools.permutations(np.hstack([np.zeros(n_stragglers),np.ones(n_workers-n_stragglers)]),n_workers)))
    # print(S)
    # S=unique_rows(S)

    S = np.ones((int(sp.binom(n_workers, n_stragglers)), n_workers))
    combs = itertools.combinations(range(n_workers), n_stragglers)
    i = 0
    for pos in combs:
        S[i, pos] = 0
        i += 1

    (m, n) = S.shape
    A = np.zeros([m, n])
    for i in range(m):
        sp_pos = S[i, :] == 1
        A[i, sp_pos] = np.linalg.lstsq(B[sp_pos, :].T, np.ones(n_workers))[0]

    return A


def compare(a, b):
    for id in range(len(a)):
        if a[id] and (not b[id]):
            return 1
        if (not a[id]) and b[id]:
            return -1
    return 0


def binary_search_row_wise(Aindex, completed, st, nd):
    if st >= nd - 1:
        return st
    idx = (st + nd) / 2
    cp = compare(Aindex[idx, :], completed)
    if cp == 0:
        return idx
    elif cp == 1:
        return binary_search_row_wise(Aindex, completed, st, idx)
    else:
        return binary_search_row_wise(Aindex, completed, idx + 1, nd)


def calculate_indexA(boolvec):
    l = len(boolvec)
    ctr = 0
    ind = 0
    for j in range(l - 1, -1, -1):
        if boolvec[j]:
            ctr = ctr + 1
            ind = ind + sp.binom(l - 1 - j, ctr)

    return int(ind)


def calculate_loss(y, predy, n_samples):
    return np.mean(np.log(1 + np.exp(-np.multiply(y, predy))))  # / n_samples


def slice_node(
    array: np.ndarray,
    n_slices: int,
    input_features: int,
    output_features: int,
    slice_where: str = "out",
):
    """
    reshape a 2D array into 3D array with shape n_slices x input_features x output_features / n_slices
    """
    assert array.shape == (input_features, output_features), (
        array.shape,
        input_features,
        output_features,
    )
    if slice_where == "out":
        output = np.zeros(
            (n_slices, input_features, output_features // n_slices)
        )
        for i in range(n_slices):
            output[i, :, :] = array[
                :,
                i
                * output_features
                // n_slices : (i + 1)
                * output_features
                // n_slices,
            ]
    if slice_where == "in":
        output = np.zeros(
            (n_slices, input_features // n_slices, output_features)
        )
        for i in range(n_slices):
            output[i, :, :] = array[
                i
                * input_features
                // n_slices : (i + 1)
                * input_features
                // n_slices,
                :,
            ]
    return output


def sigmoid(x):
    """
    Compute the sigmoid function for the input here.
    """
    return 1 / (1 + np.exp(-x))


def relu(x):
    return np.maximum(0, x)


def leaky_relu(x):
    return np.maximum(0.01 * x, x)


def leaky_relu_derivative(x):
    return np.where(x > 0, 1, 0.01)


def relu_derivative(x):
    return np.heaviside(x, 0)


def sigmoid_derivative(x):
    return sigmoid(x) * (1 - sigmoid(x))


# def softmax(x):
#     """Compute softmax values for each sets of scores in x."""
#     e_x = np.exp(x - np.max(x))
#     return e_x / (1 + e_x.sum(1).reshape(-1, 1))
# def softmax(X):
#     exps = np.exp(X - np.max(X))
#     return exps / np.sum(exps)


# def softmax_derivative(x):
#     s = x.reshape(-1, 1)
#     return np.diagflat(s) - np.dot(s, s.T)


def log_(x, p, eps):
    assert x.shape == p.shape
    out = np.zeros(x.shape)
    for idx, item in enumerate(x):
        if abs(item) < eps or abs(p[idx]) < eps:
            out[idx] = 0
        else:
            out[idx] = item * np.log(p[idx])
    return out


def message_flat(g0, g1, dims_g0, dims_g1):
    assert g0.shape == dims_g0, (g0.shape, dims_g0)
    assert g1.shape == dims_g1, (g1.shape, dims_g1)

    g0_ravelled = np.zeros(dims_g0[0] * dims_g0[1])
    g1_ravelled = np.zeros(dims_g1[0] * dims_g1[1])

    for i in range(dims_g0[0]):
        for j in range(dims_g0[1]):
            g0_ravelled[i * dims_g0[1] + j] = g0[i, j]

    for i in range(dims_g1[0]):
        for j in range(dims_g1[1]):
            g1_ravelled[i * dims_g1[1] + j] = g1[i, j]

    return np.concatenate((g0_ravelled, g1_ravelled))


def message_unflat(message, dims_g0, dims_g1):
    g0_ravelled = message[: dims_g0[0] * dims_g0[1]]
    g1_ravelled = message[dims_g0[0] * dims_g0[1] :]

    g0 = np.zeros(dims_g0)
    g1 = np.zeros(dims_g1)

    for i in range(dims_g0[0]):
        for j in range(dims_g0[1]):
            g0[i, j] = g0_ravelled[i * dims_g0[1] + j]

    for i in range(dims_g1[0]):
        for j in range(dims_g1[1]):
            g1[i, j] = g1_ravelled[i * dims_g1[1] + j]

    return g0, g1
